/*
  ProgEx35b -- Random (binary) File Access Continued

  In this example lets use the Data.dat file produced by the last example that
  just contains our two Employee records of Marty and Ray.  But lets open the
  file in "rb+" mode which is read/write mode for existing files only, and try 
  some editing in terms of updates and additions.  We'll output the contents of
  the first record, then update it with a new record.  Then we'll add a third
  record onto the end of the file.  Finally we'll dump the data.

  In the code below I've added my home-made RecordCount function because the
  Standard C Library functions don't seem to have one.  Also, I've included
  wrapper functions around the fseek() / fead() and fseek() / fwrite() functions
  in terms of BASIC like Get() / Put() syntax.  

  As I mentioned, use the Data.dat file from the last example containing our
  two records for this.  If we don't use an existing file it won't work because
  we're running in "rb+" mode (file needs to exist).
*/
#include <stdio.h>
#include <string.h>

struct Employee    
{
 char szName[32];   
 char szAddress[32];
 char szCity[16];
 char szState[16];
 char szCountry[16];
 char szZipCode[16];
 unsigned int iAge;
 unsigned int iEmpNum;
 double dblSalary;   
}; 

/* ************************************************************************* */
// The C Standard Library doesn't really have a record count function for
// binary/random access flat files so one has to be made from fseek() and
// ftell().  this...
//
// fseek(fp,0,SEEK_END); 
//
// will position the file pointer at the end of the file (feof()), and fTell
// will return the byte position at that point.  The byte offset at the feof
// divided by the length of a struct being read/written will yield the count
// of records.  Note that the function preserves the file pointer position. 
/* ************************************************************************* */
long RecordCount(FILE* fp, long iRecLength)
{
 long iTmp=0, iReturn=0, iLocation=0;

 iTmp=ftell(fp);                 //Temporarily store where caller was at
 if(iTmp==-1L)                   //If err return with error code
    return iTmp;                 //Something went wrong
 iReturn=fseek(fp,0,SEEK_END);   //Seek to very end of file
 if(iReturn)                     //If can't Seek something is wrong.  Return err
    return iReturn;              //Return -1 (error code)
 iLocation=ftell(fp);            //Get byte position of EOF in iLocation
 fseek(fp,iTmp,SEEK_SET);        //Seek back to where caller was at held in iTmp

 return iLocation/iRecLength;    //return total bytes divided by record length
}


long Put(FILE* fp, long iRecordNumber, void* pUdt, long iSize)
{
 long iReturn=0;

 if(iRecordNumber)
 {
    if(fseek(fp,(iRecordNumber-1)*iSize,SEEK_SET))
       return false;
    iReturn=fwrite(pUdt,iSize,1,fp);
    if(iReturn==1)
       return true;
 }

 return false;
}


long Get(FILE* fp, long iRecordNumber, void* pUdt, long iSize)
{
 long iReturn=0;

 if(iRecordNumber)
 {
    if(fseek(fp,(iRecordNumber-1)*iSize,SEEK_SET))
       return false;
    iReturn=fread(pUdt,iSize,1,fp);
    if(iReturn==1)
       return true;
 }

 return false; 
}


int main(void)
{
 long iRecCt=0;
 FILE* fp=NULL;
 Employee emp;

 memset(&emp,0,sizeof(Employee)); //zero out emp record
 fp=fopen("Data.dat","rb+");      //rb+ read/write binary; file must exist
 if(fp)                           //test for NULL; If fopen() fails don't cont.
 {
    puts("File Apparently Opened!  Lets First Read The 1st Record\n\n");
    if(Get(fp,1,&emp,sizeof(Employee)))
    {
       puts("Rec#    Name            Address                 City            State   Country ZipCode Age     EmpNum  Salary");
       puts("=================================================================================================================");
       printf
       (
         "1\t%s\t%s\t%s\t%s\t%s\t%s\t%u\t%u\t%6.2f\n\n",
         emp.szName,
         emp.szAddress,
         emp.szCity,
         emp.szState,
         emp.szCountry,
         emp.szZipCode,
         emp.iAge,
         emp.iEmpNum,
         emp.dblSalary
       );
    }
    else
       puts("Failed To Read First Record!");
    
    //Modify 1st record (Marty out - Eddie in)
    puts("\nNow Lets Put Somebody Else In The 1st (Zeroth) Record, Add Another");
    puts("Record At The End, And Dump The Output\n\n");
    memset(&emp,0,sizeof(Employee));
    strcpy(emp.szName,"Edward Kamara");         //add employee in zeroth place
    strcpy(emp.szAddress,"414 South Rock St.");
    strcpy(emp.szCity,"Shamokin");
    strcpy(emp.szState,"PA");
    strcpy(emp.szCountry,"USA");
    strcpy(emp.szZipCode,"17872");
    emp.iAge=16;
    emp.iEmpNum=789123;
    emp.dblSalary=10000;
    if(!Put(fp,1,&emp,sizeof(Employee))) //Put() returns false on failure so If
       puts("Failed To 'Put' Record!");  //will evaluate to true if we negate
    memset(&emp,0,sizeof(Employee));     //failure

    //Add a third record
    strcpy(emp.szName,"Daniel Higgens");         //add employee in record
    strcpy(emp.szAddress,"126 North Pearl St."); //number 3
    strcpy(emp.szCity,"Shamokin");
    strcpy(emp.szState,"PA");
    strcpy(emp.szCountry,"USA");
    strcpy(emp.szZipCode,"17872");
    emp.iAge=39;
    emp.iEmpNum=192837;
    emp.dblSalary=80000;
    if(!Put(fp,3,&emp,sizeof(Employee)))
       puts("Failed To 'Put' Record!");
    memset(&emp,0,sizeof(Employee));

    //Dump data as it now exists, i.e., altered first record and added third
    iRecCt=RecordCount(fp,sizeof(Employee));
    puts("Rec#    Name            Address                 City            State   Country ZipCode Age     EmpNum  Salary");
    puts("=================================================================================================================");
    for(unsigned int i=1; i<=iRecCt; i++)
    {
        if(Get(fp,i,&emp,sizeof(Employee)))
        {
           printf
           (
             "%u\t%s\t%s\t%s\t%s\t%s\t%s\t%u\t%u\t%6.2f\n",
             i,
             emp.szName,
             emp.szAddress,
             emp.szCity,
             emp.szState,
             emp.szCountry,
             emp.szZipCode,
             emp.iAge,
             emp.iEmpNum,
             emp.dblSalary
           );
        }
    }
    fclose(fp);
 }
 else
    puts("Couldn't Open File!");
 getchar();

 return 0;
}

/*
File Apparently Opened!  Lets First Read The 1st Record


Rec#    Name            Address                 City            State   Country ZipCode Age     EmpNum  Salary
=================================================================================================================
1       Martin Veneski  1016 Wabash Street      Shamokin        PA      USA     17872   23      123456  45000.00


Now Lets Put Somebody Else In The 1st (Zeroth) Record, Add Another
Record At The End, And Dump The Output


Rec#    Name            Address                 City            State   Country ZipCode Age     EmpNum  Salary
=================================================================================================================
1       Edward Kamara   414 South Rock St.      Shamokin        PA      USA     17872   16      789123  10000.00
2       Ray Latchkoski  1624 Tioga Street       Shamokin        PA      USA     17872   25      789123  50000.00
3       Daniel Higgens  126 North Pearl St.     Shamokin        PA      USA     17872   39      192837  80000.00
*/
